/*
 * Screen space rays by rimina
 */

THREE.DepthScreen = {

	uniforms: {
		"tDepth":   { type: "t", value: null },
		"tDiffuse":   { type: "t", value: null },
		"uNear": { type: "f", value: 0 },
		"uFar":   { type: "f", value: 0 },
		"uRo": { type: "v3", value: new THREE.Vector3(0, 0, 0)},
		"uRd": { type: "v3", value: new THREE.Vector3(0, 0, 0)},
		"uTime" : { type: "f", value: 0.0},
		"uResolution" : {type: "v2", value: new THREE.Vector2(0, 0)},
		"uFOV" : {type : "f", value: 50.0}
	},

	vertexShader: [

		"varying vec2 vUv;",
		"void main() {",
			"vUv = uv;",
			"gl_Position = projectionMatrix * modelViewMatrix * vec4( position, 1.0 );",
		"}"

	].join("\n"),

	fragmentShader: [
		"#include <packing>",
        "uniform sampler2D tDepth;",
        "uniform sampler2D tDiffuse;",
        "uniform float uNear;",
		"uniform float uFar;",
		"uniform vec3 uRo;",
		"uniform vec3 uRd;",

		"uniform float uTime;",
		"uniform vec2 uResolution;",
		"uniform float uFOV;",

        "varying vec2 vUv;",

        "#define STEPS 60",
		"#define EPSILON 0.01",
		"#define FAR 200.0",
		"#define PI 3.14159265359",

		"const vec3 FOG_COLOR = vec3(0.0, 0.25, 0.25);",
		"const vec3 LIGHT_COLOR = vec3(1.0, 0.8, 0.9);",

		//Distance functions from Mercury's SDF library
		//http://mercury.sexy/hg_sdf/

		// Maximum/minumum elements of a vector
		"float vmax3(vec3 v) {",
			"return max(max(v.x, v.y), v.z);",
		"}",

		"float fSphere(vec3 p, float r) {",
			"return length(p) - r;",
		"}",

		// Box: correct distance to corners
		"float fBox(vec3 p, vec3 b) {",
			"vec3 d = abs(p) - b;",
			"return length(max(d, vec3(0))) + vmax3(min(d, vec3(0)));",
		"}",

		// Plane with normal n (n is normalized) at some distance from the origin
		"float fPlane(vec3 p, vec3 n, float distanceFromOrigin) {",
			"return dot(p, n) + distanceFromOrigin;",
		"}",

		// Repeat in two dimensions
		"vec2 pMod2(inout vec2 p, vec2 size) {",
			"vec2 c = floor((p + size*0.5)/size);",
			"p = mod(p + size*0.5,size) - size*0.5;",
			"return c;",
		"}",

		"void rot(inout vec2 p, float a){",
			"p = cos(a)*p + sin(a)*vec2(p.y, -p.x);",
		"}",

		// 3D noise function (IQ)
		"float noise(vec3 p){",
			"vec3 ip = floor(p);",
		    "p -= ip;",
		    "vec3 s = vec3(7.0,157.0,113.0);",
		    "vec4 h = vec4(0.0, s.yz, s.y+s.z)+dot(ip, s);",
		    "p = p*p*(3.0-2.0*p);",
		    "h = mix(fract(sin(h)*43758.5), fract(sin(h+s.x)*43758.5), p.x);",
		    "h.xy = mix(h.xz, h.yw, p.y);",
		    "return mix(h.x, h.y, p.z);",
		"}",

		"vec2 scene(vec3 p){",
			"vec3 pp = p;",
			
			"float n = noise((p)*0.5);",
			"for(int i = 0; i < 4; ++i){",
				"rot(pp.xy, uTime);",
				"rot(pp.xz, uTime*1.5);",
				"pp = abs(pp) - vec3(20.0);",
				"rot(pp.xy, uTime*0.5);",
			"}",
			
		    "float obj = fBox(pp, vec3(FAR, 5.0, 5.0))-n*2.0;",
			"return vec2(obj, 1.0);",
		"}",

		"vec2 march(vec3 ro, vec3 rd){",
		    "float t = 0.0;",
		    "vec3 p = ro;",
		    "for(int i = 0; i < STEPS; ++i){",
		        "float d = scene(p).x;",
		        "p += rd*d;",
		        "t += d;",
		        
		        "if(d < EPSILON || t >= FAR){",
		            "break;",
		        "}",
		    "}",
		    "p = ro+rd*t;",
    		"float id = scene(p).y;",
		    "return vec2(t, id);",
		"}",

        "float depthScale(float d){",
        	"return d*uFar;",
        "}",

        "float depth(vec2 uv){",
        	"float z = texture2D( tDepth, uv).x;",
            "float view_z = perspectiveDepthToViewZ(z, uNear, uFar);",
            "float d = viewZToPerspectiveDepth(view_z, uNear, uFar);",
            "return depthScale(d);",
        "}",

		"vec3 normals(vec3 p){",
		    "vec3 eps = vec3(EPSILON, 0.0, 0.0);",
		    "return normalize(vec3(",
		        "scene(p+eps.xyy).x-scene(p-eps.xyy).x,",
		        "scene(p+eps.yxy).x-scene(p-eps.yxy).x,",
		        "scene(p+eps.yyx).x-scene(p-eps.yyx).x",
		    "));",
		"}",


		"vec3 normals2(){",
			"float depth0 = depth(vUv);",
			"float depth1 = depth(vUv+vec2(1.0, 0.0));",
			"float depth2 = depth(vUv+vec2(0.0, 1.0));",

			"vec3 p0 = uRo + uRd*depth0;",
			"vec3 p1 = uRo + uRd*depth1;",
			"vec3 p2 = uRo + uRd*depth2;",

			"return normalize(cross(p2-p0, p1-p0));",
		"}",

        "vec3 shade(vec3 rd, vec3 ld, vec3 p, float id){",
        	"vec3 n = vec3(0.0);",
        	"vec3 color = vec3(1.0);",
        	"if( id < 0.0){",
        		"color = texture2D(tDiffuse, vUv).rgb;",
        	"}",
        	"else{",
        		"n = normals(p);",
        		"float lambertian = max(dot(n, ld), 0.0);",
			    "float angle = max(dot(reflect(ld, n), rd), 0.0);",
			    "float specular = pow(angle, 20.0);",
			    
			    "color = lambertian*LIGHT_COLOR*vec3(0.8, 0.8, 1.0)*0.5 +",
			        "specular*LIGHT_COLOR;",
        	"}",
		    "return color;",
		"}",

		//http://www.iquilezles.org/www/articles/fog/fog.htm
		"vec3 fog(vec3 col, vec3 p, vec3 ro, vec3 rd, vec3 ld){",
			"float dist = length(p-ro);",
			"float sunAmount = max( dot(rd, -ld), 0.0 );",
			"float fogAmount = 1.0 - exp( -dist*0.01);",
			"vec3  fogColor = mix(FOG_COLOR, LIGHT_COLOR, pow(sunAmount, 8.0));",
			"return mix(col, fogColor, fogAmount);",
		"}",

        "void main() {",

        	"vec2 q = -1.0+2.0*vUv;",
    		"q.x *= uResolution.x/uResolution.y;",

            "vec3 color = vec3(0.0);",
            "float d = depth(vUv);",

            "vec3 z = normalize(uRd - uRo);",
            "vec3 x = normalize(cross(z, vec3(0.0, 1.0, 0.0)));",
    		"vec3 y = normalize(cross(x, z));",
    		"vec3 rd = normalize(mat3(x, y, z)*vec3(q, 1.0/radians(uFOV)));",

            "vec2 t = march(uRo, rd);",

            "if(d < t.x){",
            	"t.x = d;",
            	"t.y = -1.0;",
            "}",

            "vec3 p = uRo + rd*t.x;",
            "vec3 ld = normalize(uRo-p);",
            "vec3 flo = vec3(0.0, 1.5, -1.0);",
            "vec3 flt = vec3(0.0, 0.0, 0.0);",
            "vec3 fld = normalize(flt-flo);",
            "if( t.x < uFar){",
    			"color = shade(rd, ld, p, t.y);",
    		"}",
    		"color = fog(color, p, uRo, rd, fld);",

            "gl_FragColor = vec4(color, 1.0);",
        "}"

	].join("\n")

};
